/**
 *  Copyright (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  Licensed under the GPL, Version 3.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/gpl.html
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * --------------------------------------------------------------------------
 *  版权 (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  本开源软件遵循 GPL 3.0 协议;
 *  如果您不遵循此协议，则不被允许使用此文件。
 *  你可以从下面的地址获取完整的协议文本
 *
 *       http://www.gnu.org/licenses/gpl.html
 */
package org.tinygroup.linklist.impl;

import org.tinygroup.commons.exceptions.NotExistException;
import org.tinygroup.commons.processor.Processor;
import org.tinygroup.linklist.CircularLinkList;
import org.tinygroup.linklist.LinkList;

/**
 * 双向循环链表(线程安全).<br>
 * User: luoguo<br>
 * Date: 11-4-2 <br>
 * Time: 上午8:30<br>
 * To change this template use File | Settings | File Templates.
 */
public class CircularLinkListImpl<E> implements LinkList<E>, CircularLinkList {
	private int size = 0;
	private Node<E> pointer = null;
	private Object so = new Object();

	private class Node<T> {
		private T element;
		private Node<T> next = null;
		private Node<T> previous = null;

		Node(T e) {
			this.element = e;
		}
	}

	/**
	 * 
	 * 在头结点之前添加元素 并指定新元素为头结点
	 * 
	 */
	public E addFirst(E o) {
		synchronized (so) {
			if (size == 0) {
				pointer = new Node<E>(o);
				pointer.next = pointer;
				pointer.previous = pointer;
			} else {
				Node<E> n = new Node<E>(o);
				pointer.previous.next = n;
				n.previous = pointer.previous;
				pointer.previous = n;
				n.next = pointer;
				pointer = n;
			}
			size++;
			return o;
		}
	}

	/**
	 * 在链表结尾添加结点 返回所添加的元素
	 */
	public E addLast(E o) {
		synchronized (so) {
			if (size == 0) {
				pointer = new Node<E>(o);
				pointer.next = pointer;
				pointer.previous = pointer;
			} else {
				Node<E> n = new Node<E>(o);
				pointer.previous.next = n;
				n.previous = pointer.previous;
				pointer.previous = n;
				n.next = pointer;
			}
			size++;
			return o;

		}
	}

	/**
	 * 获取头结点
	 */
	public E getFirst() {
		if (size == 0) {
			throw new NotExistException();
		} else {
			return pointer.element;
		}
	}

	/**
	 * 获取结尾结点
	 */
	public E getLast() {
		if (size == 0) {
			throw new NotExistException();
		} else {
			return pointer.previous.element;
		}
	}

	/**
	 * 删除头结点 返回所删除结点
	 */
	public E removeFirst() {
		synchronized (so) {
			if (size == 0) {
				throw new NotExistException();
			} else if (size == 1) {
				E e = pointer.element;
				pointer = null;
				size--;
				return e;
			} else {
				E e = pointer.element;
				Node<E> p = pointer;
				pointer.next.previous = pointer.previous;
				pointer.previous.next = pointer.next;
				pointer = p.next;
				size--;
				return e;
			}
		}
	}

	/**
	 * 删除结尾结点 返回所删除结点
	 */
	public E removeLast() {
		synchronized (so) {
			if (size == 0) {
				throw new NotExistException();
			}
			if (size == 1) {
				E e = pointer.element;
				pointer = null;
				size--;
				return e;
			} else {
				E e = pointer.previous.element;
				pointer.previous.previous.next = pointer;
				pointer.previous = pointer.previous.previous;
				size--;
				return e;
			}
		}
	}

	/**
	 * 删除指定结点 返回所删除结点
	 */
	public E remove(E e) {
		synchronized (so) {
			if (size > 0) {
				Node<E> n = findNode(e);
				if (n == null) {
					throw new NotExistException();
				}
				if (n.equals(pointer)) {
					return removeFirst();
				} else {
					E r = n.element;
					n.next.previous = n.previous;
					n.previous.next = n.next;
					n.next = null;
					n.previous = null;
					n = null;
					size--;
					return r;
				}
			}
			throw new NotExistException();
		}
	}

	private Node<E> findNode(E e) {
		if (size == 0) {
			throw new NotExistException();
		}
		Node<E> n = pointer;
		do {
			if (n.element.equals(e)) {
				return n;
			}
			n = n.next;
		} while (!n.equals(pointer));
		throw new NotExistException();
	}

	/**
	 * 清除链表所有结点
	 */
	public void clear() {
		synchronized (so) {
			pointer = null;
			size = 0;
		}
	}

	/**
	 * 返回链表长度
	 */
	public int size() {
		return size;
	}

	/**
	 * 遍历链表，使链表所有结点都作为IProcess接口方法的参数
	 */
	public void foreach(Processor<E> eiProcess) {
		Node<E> n = pointer;
		if (size == 0) {
			return;
		}
		do {
			eiProcess.process(n.element);
			n = n.next;
		} while (!n.equals(pointer));
	}

	/**
	 * 在指定结点前添加新元素 返回所添加的元素？
	 */
	public E addBefore(E find, E o) {
		synchronized (so) {
			if (size > 0) {
				Node<E> n = findNode(find);
				if (n != null) {
					Node<E> nn = new Node<E>(o);
					n.previous.next = nn;
					n.next.previous = nn;
					n.previous = nn;
					nn.next = n;
					size++;
					return o;
				}
			}
			throw new NotExistException();
		}
	}

	/**
	 * 在指定结点的后面添加新元素 返回所添加的元素
	 */
	public E addAfter(E find, E o) {
		synchronized (so) {
			if (size > 0) {
				Node<E> n = findNode(find);
				if (n != null) {
					Node<E> nn = new Node<E>(o);
					n.next.previous = nn;
					nn.next = n.next;
					n.next = nn;
					nn.previous = n;
					size++;
					return o;
				}
			}
			throw new NotExistException();
		}
	}

	/**
	 * 后移一个结点
	 */
	public void next() {
		pointer = pointer.next;
	}

	/**
	 * 前移一个结点
	 */
	public void previous() {
		pointer = pointer.previous;
	}
}
