/**
 *  Copyright (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  Licensed under the GPL, Version 3.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/gpl.html
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * --------------------------------------------------------------------------
 *  版权 (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  本开源软件遵循 GPL 3.0 协议;
 *  如果您不遵循此协议，则不被允许使用此文件。
 *  你可以从下面的地址获取完整的协议文本
 *
 *       http://www.gnu.org/licenses/gpl.html
 */
package org.tinygroup.linklist.impl;

import org.tinygroup.commons.exceptions.NotExistException;
import org.tinygroup.commons.processor.Processor;
import org.tinygroup.linklist.LinkList;

/**
 * 双向链表(线程安全类).<br>
 * User: luoguo<br>
 * Date: 11-4-2 <br>
 * Time: 上午8:30<br>
 * To change this template use File | Settings | File Templates.
 */
public class DoubleLinkList<E> implements LinkList<E> {
	private int size = 0;
	private Node<E> head = null;
	private Node<E> tail = null;
	private Object so = new Object();

	private class Node<T> {
		private T element;
		private Node<T> next = null;
		private Node<T> previous = null;

		Node(T e) {
			this.element = e;
		}
	}

	/**
	 * 在链表头结点前添加新元素 返回所添加的元素
	 * 
	 * @param o
	 * @return <E>
	 */
	public E addFirst(E o) {
		synchronized (so) {
			if (size == 0) {
				head = new Node<E>(o);
				tail = head;
			} else {
				Node<E> n = new Node<E>(o);
				n.next = head;
				head.previous = n;
				head = n;
			}
			size++;
			return o;
		}
	}

	/**
	 * 在链表尾结后添加新元素 返回所添加的元素
	 * 
	 * @param
	 * @return
	 */
	public E addLast(E o) {
		synchronized (so) {
			if (size == 0) {
				head = new Node<E>(o);
				tail = head;
			} else {
				Node<E> n = new Node<E>(o);
				tail.next = n;
				n.previous = tail;
				tail = n;
			}
			size++;
			return o;
		}
	}

	/**
	 * 获取链表头结点
	 * 
	 * @return <E>
	 */
	public E getFirst() {
		if (size == 0) {
			throw new NotExistException();
		} else {
			return head.element;
		}
	}

	/**
	 * 获取链表尾结点
	 * 
	 * @return <E>
	 */
	public E getLast() {
		if (size == 0) {
			throw new NotExistException();
		} else {
			return tail.element;
		}
	}

	/**
	 * 删除链表头结点 返回所删除的结点
	 * 
	 * @return <E>
	 */
	public E removeFirst() {
		synchronized (so) {
			if (head == null) {
				throw new NotExistException();
			} else {
				E e = head.element;
				head = head.next;
				if (head != null) {
					head.previous = null;
				}
				size--;
				return e;
			}
		}
	}

	/**
	 * 删除链表尾结点 返回所删除的结点
	 * 
	 * 
	 * @return
	 */
	public E removeLast() {
		synchronized (so) {
			if (tail == null) {
				throw new NotExistException();
			} else {
				Node<E> n = tail;
				tail = tail.previous;
				if (tail != null) {
					tail.next = null;
				}
				size--;
				return n.element;
			}
		}
	}

	/**
	 * 
	 */

	public E remove(E e) {
		synchronized (so) {
			if (size > 0) {
				Node<E> n = findNode(e);
				if (n == null) {
					throw new NotExistException();
				}
				if (n.equals(head)) {
					if (head.next != null) {
						head.next.previous = null;
					}// 防止内存泄漏
					head = head.next;
				} else if (n.equals(tail)) {
					if (tail.previous != null) {
						tail.previous.next = null;
					}
					tail = tail.previous;
				} else {
					n.previous.next = n.next;
					n.next.previous = n.previous;
				}
				size--;
				return n.element;
			}
			throw new NotExistException();
		}
	}

	private Node<E> findNode(E e) {
		Node<E> n = head;
		do {
			if (n.element.equals(e)) {
				return n;
			}
			n = n.next;
		} while (n != null);
		throw new NotExistException();
	}

	public void clear() {
		synchronized (so) {
			head = null;
			tail = null;
			size = 0;
		}
	}

	public int size() {
		return size;
	}

	public void foreach(Processor<E> eiProcess) {
		Node<E> n = head;
		while (n != null) {
			eiProcess.process(n.element);
			n = n.next;
		}
	}

	public E addBefore(E find, E o) {
		synchronized (so) {
			if (size > 0) {
				Node<E> n = findNode(find);
				if (n != null) {
					Node<E> nn = new Node<E>(o);
					if (n.equals(head)) {
						nn.next = head;
						head.previous = nn;
						head = nn;
					} else {
						nn.next = n;
						n.previous.next = nn;
						nn.previous = n.previous;
						n.previous = nn;
					}
					size++;
					return o;
				}
			}
			throw new NotExistException();
		}
	}

	public E addAfter(E find, E o) {
		synchronized (so) {
			if (size > 0) {
				Node<E> n = findNode(find);
				if (n != null) {
					Node<E> nn = new Node<E>(o);
					if (n.equals(tail)) {
						nn.previous = n;
						tail.next = nn;
						tail = nn;
					} else {
						nn.previous = n;
						n.next.previous = nn;
						nn.next = n.next;
						n.next = nn;
					}
					size++;
					return o;
				}
			}
			throw new NotExistException();
		}
	}

}
