/**
 *  Copyright (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  Licensed under the GPL, Version 3.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/gpl.html
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * --------------------------------------------------------------------------
 *  版权 (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  本开源软件遵循 GPL 3.0 协议;
 *  如果您不遵循此协议，则不被允许使用此文件。
 *  你可以从下面的地址获取完整的协议文本
 *
 *       http://www.gnu.org/licenses/gpl.html
 */
package org.tinygroup.linklist.impl;

import org.tinygroup.commons.exceptions.NotExistException;
import org.tinygroup.commons.processor.Processor;
import org.tinygroup.linklist.LinkList;

/**
 * Created by IntelliJ IDEA.mvn User: luoguo Date: 11-4-1 Time: 下午9:13 To change
 * this template use File | Settings | File Templates.
 */

/**
 * 单向链表 线程安全类
 */
public class LinkListImpl<E> implements LinkList<E> {
	private int size = 0;
	private Node<E> root = null;
	private Object so = new Object();

	private class Node<T> {
		private T element;
		private Node<T> next = null;

		Node(T e) {
			this.element = e;
		}
	}

	/**
	 * 将指定元素插入到此链表的开头 返回所插入的元素
	 * 
	 * @param e
	 * @return
	 */
	public E addFirst(E o) {
		synchronized (so) {
			if (size == 0) {
				root = new Node<E>(o);
			} else {
				Node<E> n = new Node<E>(o);
				n.next = root;
				root = n;

			}

			size++;
			return o;
		}
	}

	/**
	 * 将指定元素插入到此链表的结尾 返回所插入的元素
	 * 
	 * @param e
	 * @return
	 */
	public E addLast(E o) {
		synchronized (so) {
			if (size == 0) {
				root = new Node<E>(o);
			} else {
				Node<E> n = root;
				while (n.next != null) {
					n = n.next;
				}
				n.next = new Node<E>(o);
			}
			size++;
			return o;
		}
	}

	/**
	 * 获取链表的第一个元素
	 * 
	 * @return
	 */
	public E getFirst() {
		if (size == 0) {
			throw new NotExistException();
		} else {
			return root.element;
		}
	}

	/**
	 * 获取链表的最后一个元素
	 * 
	 * @return
	 */
	public E getLast() {
		if (size == 0) {
			throw new NotExistException();
		} else {
			Node<E> n = root;
			while (n.next != null) {
				n = n.next;
			}
			return n.element;
		}
	}

	/**
	 * 删除链表的第一个元素 返回所删除的元素
	 * 
	 * @return
	 */
	public E removeFirst() {
		synchronized (so) {
			if (root == null) {
				throw new NotExistException();
			} else {
				E e = root.element;
				root = root.next;
				size--;
				return e;
			}
		}
	}

	/**
	 * 删除链表的最后一个元素 返回所删除的元素
	 * 
	 * @return
	 */
	public E removeLast() {
		synchronized (so) {
			if (size == 0) {
				throw new NotExistException();
			} else if (size == 1) {
				E e = root.element;
				root = null;
				size--;
				return e;

			} else {
				Node<E> n = root;
				while (n.next.next != null) {
					n = n.next;
				}
				E e = n.next.element;
				n.next = null;
				size--;
				return e;
			}
		}
	}

	/**
	 * 删除指定的元素 返回所删除的元素
	 * 
	 * @param e
	 * @return @
	 */
	public E remove(E find) {
		synchronized (so) {
			if (size == 0) {
				throw new NotExistException();
			} else if (root.element.equals(find)) {
				root = root.next;
				size--;
				return find;
			} else {
				Node<E> n = root;
				while (n.next != null) {
					if (n.next.element.equals(find)) {
						n.next = n.next.next;
						size--;
						return find;
					}
					n = n.next;
				}
			}
			throw new NotExistException();
		}
	}

	/**
	 * 从些链表中清除所有的元素
	 */
	public void clear() {
		synchronized (so) {
			root = null;
			size = 0;
		}
	}

	/**
	 * 获取链表的元素个数
	 * 
	 * @return
	 */
	public int size() {
		return size;
	}

	/**
	 * 遍历链表的所有元素 使每个元素都作为IProcess接口内process()方法的参数
	 * 
	 * @param process
	 */
	public void foreach(Processor<E> eiProcess) {
		Node<E> n = root;
		while (n != null) {
			eiProcess.process(n.element);
			n = n.next;
		}
	}

	/**
	 * 在指定元素前添加元素 返回所添加的元素
	 * 
	 * @param find
	 * @param e
	 * @return @
	 */
	public E addBefore(E find, E o) {
		synchronized (so) {
			if (size == 0) {
				throw new NotExistException();
			} else if (root.element.equals(find)) {
				Node<E> n = new Node<E>(o);
				n.next = root;
				root = n;
				size++;
				return o;
			} else {
				Node<E> n = root;
				while (n.next != null) {
					if (n.next.element.equals(find)) {
						Node<E> newNode = new Node<E>(o);
						newNode.next = n.next;
						n.next = newNode;
						size++;
						return o;
					}
					n = n.next;
				}
			}
			throw new NotExistException();
		}
	}

	/**
	 * 在指定元素后添加元素 返回所添加的元素
	 * 
	 * @param find
	 * @param e
	 * @return @
	 */
	public E addAfter(E find, E o) {
		synchronized (so) {
			if (size == 0) {
				throw new NotExistException();
			} else {
				Node<E> n = root;
				while (n != null) {
					if (n.element.equals(find)) {
						Node<E> newNode = new Node<E>(o);
						newNode.next = n.next;
						n.next = newNode;
						size++;
						return o;
					} else{
						n = n.next;
					}
				}
			}
			throw new NotExistException();
		}
	}
}
